/* N-body calculation using Greenspan 1st-order discrete mechanics. */
/* Incorporates xyz-parallelism, uses domain pointers               */

#include <stdio.hs>
#include <math.hs>
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs. */
#define N 64

int NB;      /* Number of bodies in system, <= N */
float G;     /* Gravitational constant */

/* The bodies.  here is true for entries presently in use. */

domain particle { float x0, x1, m, v0, v1, ac;  int here; };

/* Pairs of particles */

domain pair { int j, act, sm;  float mj, d0, r0, ax;
	      domain particle *pi, *pj;  domain pair *r; };

domain pair P[3][N][N];

domain particle B[3][N];

/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.  'act' identifies those
   pairs used in current problem (NB bodies). 'sm' identifies those
   procs used in summation (active + pii's) */

void pair::setup()
 { int offset, i, c;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);  act = sm && (i != j);
   pi = &B[c][i];  pj = &B[c][j];  r = (c<2) ? this+N*N : this-2*N*N;
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { int k;  float a, b, c;
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      B[0][k].x0 = a;  B[1][k].x0 = b;  B[2][k].x0 = c;
    };
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      B[0][k].v0 = a;  B[1][k].v0 = b;  B[2][k].v0 = c;
    };
   for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e", &a);  B[0][k].m = B[1][k].m = B[2][k].m = a; };
   fscanf(datap, "%e", &G);
 }

void particle::print_state()
 { mono int k;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < NB; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, B[0][k].x0, B[1][k].x0, B[2][k].x0,
               B[0][k].v0, B[1][k].v0, B[2][k].v0, B[0][k].m);
 }

void particle::print_state1()
 { mono int k;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < NB; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, B[0][k].x1, B[1][k].x1, B[2][k].x1,
               B[0][k].v1, B[1][k].v1, B[2][k].v1, B[0][k].m);
 }


void particle::print_ac()
 { mono int k;
   printf("\n       ax        ay       az\n");
   for ( k = 0; k < NB; k++ )
     printf("%2d %9.5f %9.5f %9.5f\n", k, B[0][k].ac, B[1][k].ac, B[2][k].ac);
 }

/* Calculate accelerations */

void accel()
 { [domain pair].
    { float d1, r1;  int smk, smm;
      mono int k;
      ax = 0.0;
      if (act)
      { d1 = pj->x1 - pi->x1;
	r1 = d1*d1;  ax = r1 + r->r1;  r1 = sqrt(r1 + r->ax);
        ax = G*mj*(d0 + d1)/(r0*r1*(r0 + r1)); };
      /* Sum acceleration terms in time ~ O(log N) */
      smk = j;  smm = 1;
      if (sm)
       for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (! (smk & 1)) && ( j+k < NB) )
	     { ax += (this+k)->ax;  smk >>= 1; }
	else smm = 0; }   /* end for, end if (sm) ... */
      if (j == 0) pi->ac = ax;
    };
 }

/* Calculate initial accelerations */

void accel1()
 { [domain pair].
    { int smk, smm;
      mono int k;
      ax = 0.0;
      if (act)
      { d0 = pj->x0 - pi->x0;
	r0 = d0*d0;  ax = r0 + r->r0;  r0 = sqrt(r0 + r->ax);
        ax = G*mj*d0/(r0*r0*r0); };  /* end if (act) */
      /* Sum acceleration terms in time ~ O(log N) */
      smk = j;  smm = 1;
      if (sm)
       for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (! (smk & 1)) && ( j+k < NB) )
	     { ax += (this+k)->ax;  smk >>= 1; }
	else smm = 0; }   /* end for, end if (sm) ... */
      if (j == 0) pi->ac = ax;
    };
 }


/* Greenspan's 1st-order method.  Performs ns steps of time dt
   with solution tolerance eps.  Returns number of evaluations.
   Leaves number of iterations which were stopped at limit in cfail.
   Some printing for debugging is provided in comments.  */

int cfail;   /* Counts number of iterations that failed to converge */

int green1(ns, dt, eps)
  int ns;  float dt, eps;
 { int k, ic, v, lim=10;  float hdt, mx, my;
   v = 0;  hdt = 0.5*dt;
   [domain pair].{ mj = pj->m; };
   [domain particle].
    if (here)
    { float nx;
      for (k = 0; k < ns; k++)
       { x1 = x0;  v1 = v0;  ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do { v++;  /* print_ac(); */
	      nx = v0 + dt*ac;  mx = (>?= fabs(nx - v1));
	      v1 = nx;  /* printf("mx = %12.6e", mx); */
	      nx = x0 + hdt*(v0 + v1);
	      my = (>?= fabs(nx - x1));  /* printf("  my = %12.6e", my); */
	      if (my > mx) mx = my;
	      x1 = nx;  ic++;  /* print_state1(); */
	      /* printf("mx = %12.6e", mx);*/
	      if ( ! ( mx > eps && ic < lim ) ) break;
	      accel(); }
	 while (1);
         x0 = x1;  v0 = v1;  if (ic == lim) cfail++;
       } ;   /* end for */
    } ;
   return(v);
 }

long clock(void);

void main()
 { int ns, np, k, vl;  float dt, eps;  long tm;
   char fnm[50];
   /* Performs np integrations, each ns steps of dt */
    printf("Data file> ");  scanf("%s", fnm);
   datap = fopen(fnm, "r");
   fscanf(datap, "%d", &NB);
   [domain pair].{ setup(); };
   init();
   fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
   fclose(datap);
   printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n", G, dt, eps);
   printf("\n time = 0");
   [domain particle].{ here = ((this - &B[0][0])%N) < NB;  print_state(); };
   for ( k = 0; k < np; k++ )
    { CM_start_timer(1);
      cfail = 0;
      vl = green1(ns, dt, eps);
      CM_stop_timer(1);  CM_reset_timer();
      printf(
	"\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	 (k + 1)*ns*dt, vl, cfail);
      [domain particle].{ print_state(); };
    };
 }

/* Arrangement of data file:

    NB			Number of bodies

     x      y      z
     .      .      .     Initial positions
     .      .      .     for N points
     .      .      .

    vx     vy     vz
     .      .      .     Initial velocities
     .      .      .
     .      .      .

     m
     .                   Masses
     .
     .

     G                   Gravitational constant

    ns     np     dt     Integration parameters  */

